#!/usr/local/bin/perl

######################################
# Jeremy Gore                        #
# 565-97-7242                        #
# Final Project, Bioinformatics      #
# Program for comparing two seqences #
######################################

##################
#Key variables

## Standard set of amino acid w/hash for array values
$aa = "ARNDCQEGHILKMFPSTWYV";
%aa = ("A",0,"R",1,"N",2,"D",3,"C",4,"Q",5,"E",6,"G",7,"H",8,"I",9,"L",
       10,"K",11,"M",12,"F",13,"P",14,"S",15,"T",16,"W",17,"Y",18,"V",19);

## Standard set of nucleic acids w/hash for array values
$na = "CGAT";
%na = ("C",0,"G",1,"A",2,"T",3);

## BLOSUM matrix, from the handouts.  
@blosum = ([ 4,-1,-2,-2, 0,-1,-1, 0,-2,-1,-1,-1,-1,-2,-1, 1, 0,-3,-2, 0],
           [-1, 5, 0,-2,-3, 1, 0,-2, 0,-3,-2, 2,-1,-3,-2,-1,-1,-3,-2,-3],
           [-2, 0, 6, 1,-3, 0, 0, 0, 1,-3,-3, 0,-2,-3,-2, 1, 0,-4,-3,-3],
           [-2,-2, 1, 6,-3, 0, 2,-1,-1,-3,-4,-1,-3,-3,-1, 0,-1,-4,-3,-3],
           [ 0,-3,-3,-3, 8,-3,-4,-3,-3,-1,-1,-3,-1,-2,-3,-1,-1,-2,-2,-1],
           [-1, 1, 0, 0,-3, 5, 2,-2, 0,-3,-2, 1, 0,-3,-1, 0,-1,-2,-1,-2],
           [-1, 0, 0, 2,-4, 2, 5,-2, 0,-3,-3, 1,-2,-3,-1, 0,-1,-3,-2,-2],
           [ 0,-2, 0,-1,-3,-2,-2, 6,-2,-4,-4,-2,-3,-3,-2, 0,-2,-2,-3,-3],
           [-2, 0, 1,-1,-3, 0, 0,-2, 7,-3,-3,-1,-2,-1,-2,-1,-2,-2, 2,-3],
           [-1,-3,-3,-3,-1,-3,-3,-4,-3, 4, 2,-3, 1, 0,-3,-2,-1,-3,-1, 3],
           [-1,-2,-3,-4,-1,-2,-3,-4,-3, 2, 4,-2, 2, 0,-3,-2,-1,-2,-1, 1],
           [-1, 2, 0,-1,-3, 1, 1,-2,-1,-3,-2, 5,-1,-3,-1, 0,-1,-3,-2,-2],
           [-1,-1,-2,-3,-1, 0,-2,-3,-2, 1, 2,-1, 5, 0,-2,-1,-1,-1,-1, 1],
           [-2,-3,-3,-3,-2,-3,-3,-3,-1, 0, 0,-3, 0, 6,-4,-2,-2, 1, 3,-1],
           [-1,-2,-2,-1,-3,-1,-1,-2,-2,-3,-3,-1,-2,-4, 6,-1,-1,-4,-3,-2],
           [ 1,-1, 1, 0,-1, 0, 0, 0,-1,-2,-2, 0,-1,-2,-1, 4, 1,-3,-2,-2],
           [ 0,-1, 0,-1,-1,-1,-1,-2,-2,-1,-1,-1,-1,-2,-1, 1, 5,-2,-2, 0],
           [-3,-3,-4,-4,-2,-2,-3,-2,-2,-3,-2,-3,-1, 1,-4,-3,-2,10, 2,-3],
           [-2,-2,-2,-3,-2,-1,-2,-3, 2,-1,-1,-2,-1, 3,-3,-2,-2, 2, 6,-1],
           [ 0,-3,-3,-3,-1,-2,-2,-3,-3, 3, 1,-2, 1,-1,-2,-2, 0,-3,-1, 4]);

####################
#incidental variables
 
## determines the width of array elements for printing
$max_length=4; 

####################
#Sample values

## demonstration peptides
$d_pept1 = "ABCNYRQCLCRPM"; 
$d_pept2 = "AYCYNRCKCRBP";  

## default gap values
$gap_open = 0.5;
$gap_ext = 0;

## default value to determine progress on long sum matrix computations.  
$large = 0;

####################
#Global Program

print "\n";
print "Welcome to Jeremy Gore's MB\&B 452a project,\n"; 
print "Protein and Nucleic Acid Sequence Comparison Program\n\n";
print "Do you wish to run a [D]emo, [Y]our own comparison, or [E]xit? : ";
&intro;
print "Thank you for using this program.  Have a good one!\n";

####################
#Demonstration and User defined subroutines

########
## Starts the program subroutines
sub intro {
$ans = <STDIN>;
	if ($ans =~ /^[dD]/) {
		print "\n\n";
		&demo;
	}
	elsif ($ans =~ /^[yY]/) {
		print "\n\n";
		&user_input;
	}
	elsif ($ans =~ /^[eE]/) {
		print "\n";
	}
	else {
		print "Please enter either [D]emo, [Y]our own comparison, or [E]xit: ";
		&intro;
	}
}

########
## The demonstration
sub demo {
print "(Please hit return when the demo pauses.)\n\n";
print "This program is designed to compare two protein or nucleic acid sequences \n";
print "based on the users criteria.  As a demonstration, we will be using these \n";
print "two protein sequences from the notes:\n\n";
	$pept1 = &cleanup_seq($d_pept1);
	$pept2 = &cleanup_seq($d_pept2);
	print $pept1." and ".$pept2."\n\n";
	<STDIN>;
print "First, we determine a similarity matrix.  There are many ways to do this \n";
print "using different substitution rules.  In this program, you have the option \n";
print "of choosing two different matrices: An identity matrix, or a BLOSUM \n";
print "substitution matrix taken from the packet (for proteins only of course). \n";
print "That matrix is as follows: \n\n";
	$max_length=3;
	&print_array ($aa,$aa,@blosum);
	<STDIN>;
print "Of course any other matrix could be added and used with this program.  For \n";
print "this demonstration we will use the identity matrix: \n\n";
	@aa_ident = &make_identity($aa);
	&print_array ($aa,$aa,@aa_ident);
	<STDIN>;
print "Every amino acid in the two seqences are cross referenced against the array \n";
print "to determine the similarity between them, producing a similarity matrix: \n\n";
	$max_length=4;
	@subst = @aa_ident;
	$type = %aa;
	@similarity = &make_similarity($pept1,$pept2,@subst);
	&print_array ($pept1,$pept2,@similarity);
	<STDIN>;
print "Then the similarity matrix is used to compute a sum matrix.  Here you can \n";
print "choose a gap opening penalty and gap extending penalty.  We'll use a value \n";
print "of $gap_open for opening the gap and $gap_ext for extending. \n\n";
	@sum = &make_sum(@similarity);
	&print_array ($pept1,$pept2,@sum);
	<STDIN>;
print "Next we find the traceback.\n";
@trace = &traceback($pept1,$pept2,@sum);
print "Once the traceback is complete, we have the following alignment:\n\n";
&print_alignment(@trace);
print "The program is not restricted by sequence length, but it will slow down. \n\n";
print "This completes the demonstration.\n";
print "Do you wish to run the [D]emo again, [Y]our own comparison, or [E]xit? : ";
&intro;
}

########
## Allows the user to input values
sub user_input {
print "Do you want to compare [P]rotein or [N]ucleic Acid sequences? ";
($type,%type) = &get_type;
$l = length ($type);
	if ($l == 20) {
		print "Do you want to use the [I]dentity or [B]LOSUM similarity matrix? : ";
		@subst = &get_subst;
	}
	else {
		@subst = &make_identity($na);
	}
print "Please enter the first sequence: \n";
	$seq1 = <STDIN>;
	chop ($seq1);
	$seq1 = &cleanup_seq($seq1);
print "Please enter the second sequence: \n";
	$seq2 = <STDIN>;
	chop ($seq2);
	$seq2 = &cleanup_seq($seq2);
print "What do you want the gap opening penalty to be? : ";
	$gap_open = <STDIN>;
print "\nWhat do you want the gap extension penalty to be? : ";
	$gap_ext = <STDIN>;
print "\nThe alignment is as follows: \n\n";
@similarity = &make_similarity($seq1,$seq2,@subst);
@sum = &make_sum(@similarity);
@trace = &traceback($seq1,$seq2,@sum);
&print_alignment(@trace);
}

########
## Determines what type of substitution array to use for proteins
sub get_subst {
$ans = <STDIN>;
	if ($ans =~ /^[iI]/) {
		print "\n";
		@aa_ident = &make_identity($aa);
		@ans = @aa_ident;
	}
	elsif ($ans =~ /^[bB]/) {
		print "\n\n";
		@ans = @blosum;
	}
	else {
		print "Please enter either [I]dentity or [B]LOSUM: ";
		@ans = &get_subst;
	}
@ans;
}

########
## Determines what type of sequence to be compared
sub get_type {
$ans = <STDIN>;
	if ($ans =~ /^[pP]/) {
		$ans = $aa;
		%ans = %aa;
	}
	elsif ($ans =~ /^[nN]/) {
		$ans = $na;
		%ans = %na;
	}
	else {
		print "Please enter either [P]rotein or [N]ucleic Acid: ";
		($ans,%ans) = &get_type;
	}
($ans,%ans);
}




####################
## Key Functions
## These are the nuts and bolts of sequence comparison.  

########
## Makes a similarity matrix from sequences and substitution matrices
sub make_similarity {
$seq1 = shift @_;
$seq2 = shift @_;
@index = @_;
@seq1 = &listify($seq1);
@seq2 = &listify($seq2);
	for $i (0 .. $#seq1) {
		for $j (0 .. $#seq2) {
			$similarity[$i][$j] = $index[$aa{"$seq1[$i]"}][$aa{"$seq2[$j]"}];
		}
	}
@similarity;
}

########
## Makes a sum matrix
sub make_sum {
@sum = @_;
$x = $#sum;
$y = $#{$sum[0]};
if (($x + $x) > 200) {$large = 1}
	for ($i = $x; $i >= 0; $i-- ) {
		for ($j = $y; $j >= 0; $j-- ) {
			($max,$max_i,$max_j) = &best_next($i,$j,$gap_open,$gap_ext,@sum);
			$sum[$i][$j] = $sum[$i][$j]+$max;
		}
if ($large) {print $i/$#sum*100 . "% of sum left to compute\n"}
	}
@sum;
}

########
## Traces back through the array and makes an alignment
sub traceback {
$seq1 = shift @_;
$seq2 = shift @_;
@sum = @_;
@seq1 = &listify($seq1);
@seq2 = &listify($seq2);
($max,$a,$b) = &best_of_array(@sum);
$e = 0;
$x = 0;
$y = 0;
if ($a != $x) {
	$trace[0][$e] = $seq1[$x];
	$trace[1][$e] = " ";	
	$trace[2][$e] = "-";
			++$e;
}
if ($b != $y) {
	$trace[0][$e] = "-";
	$trace[1][$e] = " ";
	$trace[2][$e] = $seq2[$y];
			++$e;
}
while ($x <= $#sum) {
		while ($a > $x+1) {
			$trace[0][$e] = $seq1[$x+1];
			$trace[1][$e] = " ";
			$trace[2][$e] = "-";
			++$x;
			++$e;
		}
		while ($b > $y+1) {
			$trace[0][$e] = "-";
			$trace[1][$e] = " ";
			$trace[2][$e] = $seq2[$y+1];
			++$y;
			++$e;
		}
if ($a <= $#sum) {
	$trace[0][$e] = $seq1[$a];
	$trace[1][$e] = &parse_sim($seq1[$a],$seq2[$b]);
	$trace[2][$e] = $seq2[$b];
	++$e;
}
	$x = $a;
	$y = $b;
	($max,$a,$b) = &best_next($x,$y,0,0,@sum);
	}
@trace;
}



########
## Finds the largest value in the array
sub best_of_array {
@sum = @_;
$best = $sum[0][0];
$best_i = 0;
$best_j = 0;
	for $i (0 .. $#sum) {
		for $j (0 .. $#{$sum[$i]}) {
			if ($sum[$i][$j] > $best) {
				$best = $sum[$i][$j];
				$best_i = $i;
				$best_j = $j;
			}
		}
	}
($best,$best_i,$best_j);
}

########
## Finds the best value of the diagonal, row, and column down and to the right
## from the location in the array, based on gap penalties, for sum and traceback
sub best_next {
$m = shift @_;
$n = shift @_;
$g_open = shift @_;
$g_ext = shift @_;
@sum = @_;
$w = $#sum;
$z = $#{$sum[0]};
	if ($n+1 <= $z and $m+1 <= $w) {
		$best = $sum[$m+1][$n+1];
		$best_i = $m+1;
		$best_j = $n+1;
		if ($n+2 <= $z) {
			$c = $m+1;
			for $d ($n+2 .. $z) {
				$t = $sum[$c][$d]-($g_open+(($d - $n - 1)*$g_ext));
				if  ($t > $best) {
					$best = $t;
					$best_i = $c;
					$best_j = $d;
				}
			}
		}
		if ($m+2 <= $w) {
			$d = $n+1;
			for $c ($m+2 .. $w) {
				$t = $sum[$c][$d]-($g_open+(($c - $m - 1)*$g_ext));
				if ($t > $best) {
					$best = $t;
					$best_i = $c;
					$best_j = $d;
				}
			}
		}
	}
	else {
		$best = 0;
		$best_i = $w+1;
		$best_j = $z+1;
	}
($best,$best_i,$best_j);
}

########
## Determines whether corresponding acids are identical for traceback
sub parse_sim {
$some1 = shift @_;
$some2 = shift @_;
if ($some1 == $some2){$r = "|"}
else {$r = " "}
$r;
}


####################
## Maintenance functions

########
## Prints the alignment.  Could be improved for longer sequences with difficulty.   
sub print_alignment {
	for $i ( 0 .. 2 ) {
   for $j ( 0 .. $#{$_[$i]} ) {
      print "$_[$i][$j]";
        }
			print "\n"
    }
print "\n";
}

########
## Prints an array with labels for demo   
sub print_array {
$aa1 = shift @_;
$aa2 = shift @_;
@array = @_;
@aa1 = &listify($aa1);
@aa2 = &listify($aa2);
print "  ";
for $j (0 .. $#aa2) {
print " "x($max_length - 1);
print "$aa2[$j]";
}
print "\n";
for $i ( 0 .. $#aa1 ) {
	print "$aa1[$i] ";
	for $j ( 0 .. $#aa2 ) {
		$length = length ($array[$i][$j]);
  print " "x($max_length-$length);
		print "$array[$i][$j]";
		}
	print "\n";
	}
print "\n";
}

########
## Turns a string into a list   
sub listify {
@list = split (//,$_[0]);
}

########
## Cleans up the input sequences for accurate comparison.     
sub cleanup_seq {
$p = $_[0];
$p =~ s/[^A-Za-z]//g;
$p =~ tr/a-z/A-Z/;
$p =~ tr/U/T/;
$p;
}

########
## Makes an identity matrix.  Easier and cheaper than inputing the whole thing.  
sub make_identity {
$length = length($_[0])-1;
for $i (0 .. $length) {
	for $j (0 .. $length) { 
		if ($i == $j) {$e = 1;} else {$e = 0;};
		$identity[$i][$j] = $e;
		}
	}
@identity;
}